import requests

from .base import Request
from .response import Response


class StandardRequest(Request):
    """
    Performs plain HTTP requests using the `requests` library.
    """

    def __init__(self):
        self._session = requests.Session()

    def perform(self, method, url, **kwargs):
        return Response(self._session.request(method, url, **kwargs))



from .base import API
from .routing import Router


class BundleAPI(API):
    """https://private.intel.amp.cisco.com/index.html#/Bundle"""

    __router, route = Router.new()

    @route('export.get')
    def _perform(self, **kwargs):
        return self._get(
            '/ctia/bundle/export',
            **kwargs
        )

    @route('export.post')
    def _perform(self, payload, **kwargs):
        return self._post(
            '/ctia/bundle/export',
            json=payload,
            **kwargs
        )

    @route('import_.post')
    def _perform(self, payload, **kwargs):
        return self._post(
            '/ctia/bundle/import',
            json=payload,
            **kwargs
        )



from .entity import IntelEntityAPI
from .routing import Router
from .. import urls


class VulnerabilityAPI(IntelEntityAPI):
    """https://private.intel.amp.cisco.com/index.html#/Vulnerability"""

    __router, route = Router.new()

    def __init__(self, request):
        super(VulnerabilityAPI, self).__init__(request, '/ctia/vulnerability')

    @route('cpe_match_strings')
    def _perform(self, **kwargs):
        return self._get(
            urls.join(self._url, 'cpe_match_strings'),
            **kwargs
        )



# Make the classes below importable from the `.api` subpackage directly.
from .enrich import EnrichAPI
from .inspect import InspectAPI
from .response import ResponseAPI
from .intel import IntelAPI
from .commands import CommandsAPI



from six.moves.urllib.parse import quote

from .exceptions import RegionError

_url_patterns_by_api_family = {
    'visibility': 'https://visibility{region}.amp.cisco.com',
    'private_intel': 'https://private.intel{region}.amp.cisco.com',
    'global_intel': 'https://intel{region}.amp.cisco.com'
}


def _url_for_region(url_pattern, region):
    # Fall back to the default region.
    if region == 'us':
        region = ''

    return url_pattern.format(region='.' + region if region != '' else '')


def _urls_by_region(urls):
    return dict(
        (
            region,
            dict(
                (api_family, _url_for_region(url_pattern, region))
                for api_family, url_pattern in urls.items()
            )
        )
        for region in (
            '',
            'us',
            'eu',
            'apjc',
        )
    )


def url_for(region, family, environment=None):
    # Fall back to the default region.
    if region is None:
        region = ''
    # Fall back to the default environment.
    if environment is None:
        environment = _url_patterns_by_api_family
    if region not in _urls_by_region(environment):
        # Use `repr` to make each region enclosed in quotes.
        raise RegionError(
            'Invalid region {}, must be one of: {}.'.format(
                repr(region),
                ', '.join(map(repr, _urls_by_region(environment).keys())),
            )
        )

    return _urls_by_region(environment)[region][family]


def join(base, *parts):
    return base.rstrip('/') + '/' + '/'.join(
        quote(str(part).strip('/'), safe='')
        for part in parts
    )



from .asset import AssetAPI
from .base import API
from .bundle import BundleAPI
from .casebook import CasebookAPI
from .entity import IntelEntityAPI
from .event import EventAPI
from .incident import IncidentAPI
from .indicator import IndicatorAPI
from .judgement import JudgementAPI
from .routing import Router
from .sighting import SightingAPI
from .feed import FeedAPI
from .vulnerability import VulnerabilityAPI
from .. import urls


class IntelAPI(API):

    __router, route = Router.new()

    def __init__(self, request):
        super(IntelAPI, self).__init__(request)

        self._actor = IntelEntityAPI(request, '/ctia/actor')
        self._actor.__doc__ = \
            ("https://private.intel.amp.cisco.com/index.html#/Actor"
             "https://intel.amp.cisco.com/index.html#/Actor")

        self._campaign = IntelEntityAPI(request, '/ctia/campaign')
        self._campaign.__doc__ = \
            ("https://private.intel.amp.cisco.com/index.html#/Campaign"
             "https://intel.amp.cisco.com/index.html#/Campaign")

        self._coa = IntelEntityAPI(request, '/ctia/coa')
        self._coa.__doc__ = \
            ("https://private.intel.amp.cisco.com/index.html#/COA"
             "https://intel.amp.cisco.com/index.html#/COA")

        self._data_table = IntelEntityAPI(request, '/ctia/data-table')
        self._data_table.__doc__ = \
            ("https://private.intel.amp.cisco.com/index.html#/DataTable"
             "https://intel.amp.cisco.com/index.html#/DataTable")

        self._attack_pattern = IntelEntityAPI(request, '/ctia/attack-pattern')
        self._attack_pattern.__doc__ = \
            ("https://private.intel.amp.cisco.com/index.html#/Attack_Pattern"
             "https://intel.amp.cisco.com/index.html#/Attack_Pattern")

        self._feedback = IntelEntityAPI(request, '/ctia/feedback')
        self._feedback.__doc__ = \
            ("https://private.intel.amp.cisco.com/index.html#/Feedback"
             "https://intel.amp.cisco.com/index.html#/Feedback")

        self._graphql = IntelEntityAPI(request, '/ctia/graphql')
        self._graphql.__doc__ = \
            ("https://private.intel.amp.cisco.com/index.html#/GraphQL"
             "https://intel.amp.cisco.com/index.html#/GraphQL")

        self._bulk = IntelEntityAPI(request, '/ctia/bulk')
        self._bulk.__doc__ = \
            ("https://private.intel.amp.cisco.com/index.html#/Bulk"
             "https://intel.amp.cisco.com/index.html#/Bulk")

        self._malware = IntelEntityAPI(request, '/ctia/malware')
        self._malware.__doc__ = \
            ("https://private.intel.amp.cisco.com/index.html#/Malware"
             "https://intel.amp.cisco.com/index.html#/Malware")

        self._relationship = IntelEntityAPI(request, '/ctia/relationship')
        self._relationship.__doc__ = \
            ("https://private.intel.amp.cisco.com/index.html#/Relationship"
             "https://intel.amp.cisco.com/index.html#/Relationship")

        self._tool = IntelEntityAPI(request, '/ctia/tool')
        self._tool.__doc__ = \
            ("https://private.intel.amp.cisco.com/index.html#/Tool"
             "https://intel.amp.cisco.com/index.html#/Tool")

        self._investigation = IntelEntityAPI(request, '/ctia/investigation')
        self._investigation.__doc__ = \
            ("https://private.intel.amp.cisco.com/index.html#/Investigation"
             "https://intel.amp.cisco.com/index.html#/Investigation")

        self._weakness = IntelEntityAPI(request, '/ctia/weakness')
        self._weakness.__doc__ = \
            ("https://private.intel.amp.cisco.com/index.html#/Weakness"
             "https://intel.amp.cisco.com/index.html#/Weakness")

        self._identity_assertion = \
            IntelEntityAPI(request, '/ctia/identity-assertion')
        self._identity_assertion.__doc__ = \
            ("https://private.intel.amp.cisco.com/index.html#"
             "/IdentityAssertion"
             "https://intel.amp.cisco.com/index.html#/IdentityAssertion")

        self._bundle = BundleAPI(request)
        self._event = EventAPI(request)
        self._incident = IncidentAPI(request)
        self._indicator = IndicatorAPI(request)
        self._judgement = JudgementAPI(request)
        self._casebook = CasebookAPI(request)
        self._sighting = SightingAPI(request)
        self._feed = FeedAPI(request)
        self._vulnerability = VulnerabilityAPI(request)

    @property
    def actor(self):
        return self._actor

    @property
    def bundle(self):
        return self._bundle

    @property
    def campaign(self):
        return self._campaign

    @property
    def coa(self):
        return self._coa

    @property
    def data_table(self):
        return self._data_table

    @property
    def attack_pattern(self):
        return self._attack_pattern

    @property
    def feedback(self):
        return self._feedback

    @property
    def feed(self):
        return self._feed

    @property
    def graphql(self):
        return self._graphql

    @property
    def event(self):
        return self._event

    @property
    def identity_assertion(self):
        return self._identity_assertion

    @property
    def incident(self):
        return self._incident

    @property
    def indicator(self):
        return self._indicator

    @property
    def judgement(self):
        return self._judgement

    @property
    def casebook(self):
        return self._casebook

    @property
    def sighting(self):
        return self._sighting

    @property
    def bulk(self):
        return self._bulk

    @property
    def malware(self):
        return self._malware

    @property
    def relationship(self):
        return self._relationship

    @property
    def tool(self):
        return self._tool

    @property
    def investigation(self):
        return self._investigation

    @property
    def vulnerability(self):
        return self._vulnerability

    @property
    def weakness(self):
        return self._weakness

    @route('properties.get')
    def _perform(self, **kwargs):
        """
        https://private.intel.amp.cisco.com/index.html#/Properties
        https://intel.amp.cisco.com/index.html#/Properties
        """

        return self._get('/ctia/properties', **kwargs)

    @route('metrics.get')
    def _perform(self, **kwargs):
        """
        https://private.intel.amp.cisco.com/index.html#/Metrics
        https://intel.amp.cisco.com/index.html#/Metrics
        """

        return self._get('/ctia/metrics', **kwargs)

    @route('verdict.get')
    def _perform(self,
                 observable_type,
                 observable_value,
                 **kwargs):
        """
        https://private.intel.amp.cisco.com/index.html#/Verdict
        https://intel.amp.cisco.com/index.html#/Verdict
        """

        return self._get(
            urls.join('/ctia', observable_type, observable_value, 'verdict'),
            **kwargs
        )

    @route('status.get')
    def _perform(self, **kwargs):
        """
        https://private.intel.amp.cisco.com/index.html#/Status
        https://intel.amp.cisco.com/index.html#/Status
        """

        return self._get('/ctia/status', **kwargs)

    @route('version.get')
    def _perform(self, **kwargs):
        """
        https://private.intel.amp.cisco.com/index.html#/Version
        https://intel.amp.cisco.com/index.html#/Version
        """

        return self._get('/ctia/version', **kwargs)


class PrivateIntel(IntelAPI):
    """https://private.intel.amp.cisco.com/index.html"""

    def __init__(self, request):
        super(PrivateIntel, self).__init__(request)

        self._asset = IntelEntityAPI(request, '/ctia/asset')
        self._asset.__doc__ = \
            "https://private.intel.amp.cisco.com/index.html#/Asset"

        self._asset_mapping = AssetAPI(request, '/ctia/asset-mapping')
        self._asset_mapping.__doc__ = \
            "https://private.intel.amp.cisco.com/index.html#/Asset%20Mapping"

        self._asset_properties = AssetAPI(request, '/ctia/asset-properties')
        self._asset_properties.__doc__ = \
            "https://private.intel.amp.cisco.com/index.html#/" \
            "Asset%20Properties"

        self._target_record = IntelEntityAPI(request, '/ctia/target-record')
        self._target_record.__doc__ = \
            "https://private.intel.amp.cisco.com/index.html#/Target%20Record"

    @property
    def asset(self):
        return self._asset

    @property
    def asset_mapping(self):
        return self._asset_mapping

    @property
    def asset_properties(self):
        return self._asset_properties

    @property
    def target_record(self):
        return self._target_record


class GlobalIntel(IntelAPI):
    """https://intel.amp.cisco.com/index.html"""

    def __init__(self, request):
        super(GlobalIntel, self).__init__(request)



from .routing import Resolution, Router
from ..exceptions import ResponseTypeError


class API(object):
    """ Base `API`. """

    def __init__(self, request):
        self._request = request
        self._resolution = None

    def _get(self, *args, **kwargs):
        return self.__perform('GET', *args, **kwargs)

    def _post(self, *args, **kwargs):
        return self.__perform('POST', *args, **kwargs)

    def _put(self, *args, **kwargs):
        return self.__perform('PUT', *args, **kwargs)

    def _patch(self, *args, **kwargs):
        return self.__perform('PATCH', *args, **kwargs)

    def _delete(self, *args, **kwargs):
        return self.__perform('DELETE', *args, **kwargs)

    def __perform(self, method, *args, **kwargs):
        response_types = {
            'raw': lambda response: response,
            'json': lambda response: response.json(),
            'text': lambda response: response.text,
        }
        response_type = kwargs.pop('response_type', 'json')

        if response_type not in response_types:
            raise ResponseTypeError(
                'Unsupported response type {type}, must be one of:'
                ' {types}.'.format(
                    type=repr(response_type),
                    types=', '.join(map(repr, response_types.keys())),
                )
            )

        response = self._request.perform(method, *args, **kwargs)
        response.raise_for_status()

        processed = response_types[response_type]

        return processed(response)

    def __getattr__(self, item):
        if self._resolution is None:
            self._resolution = self._build_resolution()

        return self._resolution.__getattr__(item)

    def _build_resolution(self):
        """ Traverses the MRO and merges values of
        `__router` attributes to build a single `Resolution`. """

        router = None

        for cls in type(self).mro():
            attribute = '_{class_name}__{router}'.format(
                class_name=cls.__name__,
                router='router'
            )

            if hasattr(cls, attribute):
                router = Router.merged(router, getattr(cls, attribute))

        if router is None:
            raise Exception(
                'Could not build a resolution for {type}.'.format(
                    type=type(self)
                )
            )

        return Resolution(self, router)



from six.moves.http_client import UNAUTHORIZED
from six.moves.urllib.parse import urljoin

from .base import Request
from ..urls import url_for


class ClientAuthorizedRequest(Request):
    """
    Provides authorization header for inner request.
    """

    def __init__(self, request, client_id,
                 client_password, region=None, environment=None):
        self._request = request
        self._client_id = client_id
        self._client_password = client_password

        self._token_url = urljoin(
            url_for(region, 'visibility', environment),
            '/iroh/oauth2/token'
        )

        self._token = self._request_token()

    def perform(self, method, url, **kwargs):
        headers = kwargs.pop('headers', {})

        response = self._perform(method, url, headers, **kwargs)

        if response.status_code == UNAUTHORIZED:
            # The token has already expired (most probably),
            # so regenerate it again and try one more time
            self._token = self._request_token()
            response = self._perform(method, url, headers, **kwargs)

        return response

    def _request_token(self):
        data = {'grant_type': 'client_credentials'}
        headers = {'Content-Type': 'application/x-www-form-urlencoded',
                   'Accept': 'application/json'}
        auth = (self._client_id, self._client_password)  # HTTP Basic Auth

        response = self._request.post(self._token_url,
                                      data=data,
                                      headers=headers,
                                      auth=auth)

        response.raise_for_status()

        return response.json()['access_token']  # OK

    @property
    def _headers(self):
        return {'Authorization': 'Bearer {token}'.format(token=self._token)}

    def _perform(self, method, url, headers, **kwargs):
        headers.update(self._headers)
        kwargs['headers'] = headers
        return self._request.perform(method, url, **kwargs)


class TokenAuthorizedRequest(Request):
    """
    Provides authorization header for inner request.
    """

    def __init__(self, request, token, region=None, environment=None):
        self._request = request
        self._token = token
        self._check_url = urljoin(
            url_for(region, 'visibility', environment),
            '/iroh/iroh-enrich/settings',
        )
        self._check_token()

    def perform(self, method, url, **kwargs):
        headers = kwargs.pop('headers', {})
        response = self._perform(method, url, headers, **kwargs)
        return response

    def _check_token(self):
        headers = {'Accept': 'application/json'}
        headers.update(self._headers)

        response = self._perform('GET', self._check_url, headers)
        response.raise_for_status()

    @property
    def _headers(self):
        return {'Authorization': 'Bearer {token}'.format(token=self._token)}

    def _perform(self, method, url, headers, **kwargs):
        headers.update(self._headers)
        kwargs['headers'] = headers
        return self._request.perform(method, url, **kwargs)



from .base import API
from .routing import Router
from .entity import EntityAPI
from .module_type_patch import ModuleTypePatchAPI


class IntAPI(API):
    __router, route = Router.new()

    def __init__(self, request):
        super(IntAPI, self).__init__(request)

        self._integration = EntityAPI(
            request, '/iroh/iroh-int/integration')
        self._integration.__doc__ = "https://visibility.amp.cisco.com/iroh/" \
                                    "iroh-int/index.html#/Integration"

        self._module_instance = EntityAPI(
            request, '/iroh/iroh-int/module-instance')
        self._integration.__doc__ = "https://visibility.amp.cisco.com/iroh/" \
                                    "iroh-int/index.html#/ModuleInstance"

        self._module_type = EntityAPI(
            request, '/iroh/iroh-int/module-type')
        self._integration.__doc__ = "https://visibility.amp.cisco.com/iroh/" \
                                    "iroh-int/index.html#/ModuleType"

        self._module_type_patch = ModuleTypePatchAPI(request)

    @property
    def integration(self):
        return self._integration

    @property
    def module_instance(self):
        return self._module_instance

    @property
    def module_type(self):
        return self._module_type

    @property
    def module_type_patch(self):
        return self._module_type_patch



from .entity import IntelEntityAPI
from .routing import Router
from .. import urls


class CasebookAPI(IntelEntityAPI):
    """https://private.intel.amp.cisco.com/index.html#/Casebook"""

    __router, route = Router.new()

    def __init__(self, request):
        super(CasebookAPI, self).__init__(request, '/ctia/casebook')

    @route('observables')
    def _perform(self, id_, payload, **kwargs):
        return self._post(
            urls.join(self._url, id_, 'observables'),
            json=payload,
            **kwargs
        )

    @route('texts')
    def _perform(self, id_, payload, **kwargs):
        return self._post(
            urls.join(self._url, id_, 'texts'),
            json=payload,
            **kwargs
        )

    @route('bundle')
    def _perform(self, id_, payload, **kwargs):
        return self._post(
            urls.join(self._url, id_, 'bundle'),
            json=payload,
            **kwargs
        )



from .base import API
from .. import urls
from ..exceptions import ResponseTypeError


class EntityAPI(API):

    def __init__(self, request, url):
        super(EntityAPI, self).__init__(request)

        self._url = url

    def get(self, id_=None, **kwargs):
        if id_:
            url = urls.join(self._url, id_)
        else:
            url = self._url

        return self._get(url, **kwargs)

    def post(self, payload, **kwargs):
        return self._post(self._url, json=payload, **kwargs)

    def put(self, id_, payload, **kwargs):
        return self._put(
            urls.join(self._url, id_),
            json=payload,
            **kwargs
        )

    def patch(self, id_, payload, **kwargs):
        return self._patch(
            urls.join(self._url, id_),
            json=payload,
            **kwargs
        )

    def delete(self, id_, **kwargs):
        if 'response_type' in kwargs:
            raise ResponseTypeError("'response_type' cannot be "
                                    "specified for this method.")

        return self._delete(
            urls.join(self._url, id_),
            response_type='raw',
            **kwargs
        )


class Search(EntityAPI):
    NAME = 'search'

    def get(self, **kwargs):
        return self._get(
            urls.join(self._url, self.NAME),
            **kwargs
        )

    def delete(self, **kwargs):
        return self._delete(
            urls.join(self._url, self.NAME),
            **kwargs
        )

    def count(self, **kwargs):
        return self._get(
            urls.join(self._url, self.NAME, 'count'),
            **kwargs
        )


class Metric(EntityAPI):
    NAME = 'metric'

    def histogram(self, **kwargs):
        return self._get(
            urls.join(self._url, self.NAME, 'histogram'),
            **kwargs
        )

    def topn(self, **kwargs):
        return self._get(
            urls.join(self._url, self.NAME, 'topn'),
            **kwargs
        )

    def cardinality(self, **kwargs):
        return self._get(
            urls.join(self._url, self.NAME, 'cardinality'),
            **kwargs
        )


class IntelEntityAPI(EntityAPI):

    def __init__(self, request, url):
        super(IntelEntityAPI, self).__init__(request, url)
        self.search = Search(request, url)
        self.metric = Metric(request, url)

    def external_id(self, id_, **kwargs):
        return self._get(
            urls.join(self._url, 'external_id', id_),
            **kwargs
        )



import json

import requests


class Response(object):
    """
    Wraps an instance of the `requests.Response` class.
    Redirects all calls to get/set any attribute to inner response.
    May also customize some instance methods.
    """

    def __init__(self, response):
        self._response = response

    def __getattr__(self, key):
        return getattr(self._response, key)

    def __setattr__(self, key, value):
        # This is an antidote against infinite recursion:
        # in order to use self._response for redirecting calls,
        # make sure to set the '_response' attribute directly first.
        if key == '_response':
            super(Response, self).__setattr__(key, value)
        else:
            setattr(self._response, key, value)

    def raise_for_status(self):
        extended = self._extended

        try:
            self._response.raise_for_status()
        except requests.HTTPError as error:
            raise extended(error)

    @staticmethod
    def _extended(error):
        # Try to extend the default error message with the response payload
        # in order to give the user more insight about what went wrong.

        try:
            payload = json.loads(error.response.text)
        except json.JSONDecodeError:
            return error

        message = error.args[0]  # 1-element tuple.
        message += '\n' + json.dumps(payload, indent=4, sort_keys=True)

        error.args = (message,)

        return error



from .entity import IntelEntityAPI
from .routing import Router
from .. import urls


class IndicatorAPI(IntelEntityAPI):
    """https://private.intel.amp.cisco.com/index.html#/Indicator"""

    __router, route = Router.new()

    def __init__(self, request):
        super(IndicatorAPI, self).__init__(request, '/ctia/indicator')

    @route('judgements.indicators')
    def _perform(self, observable_type, observable_value, **kwargs):
        return self._get(
            urls.join(
                '/ctia',
                observable_value,
                observable_type,
                'judgements',
                'indicators'
            ),
            **kwargs
        )

    @route('sightings.indicators')
    def _perform(self, observable_type, observable_value, **kwargs):
        return self._get(
            urls.join(
                '/ctia',
                observable_type,
                observable_value,
                'sightings',
                'indicators'
            ),
            **kwargs
        )



from .entity import EntityAPI
from .. import urls


class ModuleTypePatchAPI(EntityAPI):
    """
    https://visibility.amp.cisco.com/iroh/iroh-int/index.html#/ModuleTypePatch
    """
    URL = '/iroh/iroh-int/module-type-patch'

    def __init__(self, request):
        super(ModuleTypePatchAPI, self).__init__(request, self.URL)

    def action_preview(self, id_, **kwargs):
        return self._get(
            urls.join(self._url, id_, 'action/preview'),
            **kwargs
        )



# Make the classes below importable from the `.request` subpackage directly.
from .authorized import ClientAuthorizedRequest, TokenAuthorizedRequest
from .logged import LoggedRequest
from .proxied import ProxiedRequest
from .relative import RelativeRequest
from .response import Response
from .standard import StandardRequest
from .timed import TimedRequest



from .base import API
from .routing import Router
from .. import urls


class SSEDeviceAPI(API):
    __router, route = Router.new()

    @route('get_all')
    def _perform(self, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-sse/index.html#/SSE/get_iroh_iroh_sse_device
        """

        return self._get(
            '/iroh/iroh-sse/device',
            **kwargs
        )

    @route('get_by_id')
    def _perform(self, device_id, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-sse/index.html#/SSE/get_iroh_iroh_sse_device__device_id_
        """

        return self._get(
            urls.join('/iroh/iroh-sse/device', device_id),
            **kwargs
        )

    @route('post')
    def _perform(self,
                 payload,
                 **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-sse/index.html#/SSE/post_iroh_iroh_sse_device
        """

        return self._post(
            '/iroh/iroh-sse/device',
            json=payload,
            **kwargs
        )

    @route('patch')
    def _perform(self,
                 device_id,
                 payload,
                 **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-sse/index.html#/SSE/patch_iroh_iroh_sse_device__device_id_
        """

        return self._patch(
            urls.join('/iroh/iroh-sse/device', device_id),
            json=payload,
            **kwargs
        )

    @route('token')
    def _perform(self,
                 device_id,
                 payload,
                 **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-sse/index.html#/SSE/post_iroh_iroh_sse_device__device_id__token
        """

        return self._post(
            urls.join('/iroh/iroh-sse/device', device_id, '/token'),
            json=payload,
            **kwargs
        )

    @route('re_token')
    def _perform(self,
                 device_id,
                 payload,
                 **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-sse/index.html#/SSE/post_iroh_iroh_sse_device__device_id__token
        """

        return self._post(
            urls.join('/iroh/iroh-sse/device', device_id, '/token'),
            json=payload,
            **kwargs
        )

    @route('api_proxy')
    def _perform(self,
                 device_id,
                 payload,
                 **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-sse/index.html#/SSE/post_iroh_iroh_sse_device__device_id__api_proxy
        """

        return self._post(
            urls.join('/iroh/iroh-sse/device', device_id, '/api-proxy'),
            json=payload,
            **kwargs
        )

    @route('delete')
    def _perform(self, device_id, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-sse/index.html#/SSE/delete_iroh_iroh_sse_device__device_id_
        """

        return self._delete(
            urls.join('/iroh/iroh-sse/device', device_id),
            **kwargs
        )


class SSETenantAPI(API):
    __router, route = Router.new()

    @route('get_token')
    def _perform(self, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-sse/index.html#/SSE/get_iroh_iroh_sse_tenant_token
        """

        return self._get(
            '/iroh/iroh-sse/tenant/token',
            **kwargs
        )



from itertools import chain

from ...exceptions import RouteError


class Router(object):
    """ Represents a mapping from a route to a method. """

    def __init__(self, routes=None):
        self._routes = routes or {}

    def register(self, route):
        """ Returns a function to register a method by a specified `route`. """

        def register(method):
            if route in self._routes:
                raise RouteError(
                    'Route {route} has already been registered.'.format(
                        route=repr(route)
                    )
                )

            self._routes[route] = method

            # We do not return anything
            # to set the decorated method to `None`.
            return None

        return register

    def resolve(self, route):
        """ Returns a method by the specified route. """

        if not route:
            raise RouteError('Route cannot be empty.')
        if route not in self._routes:
            raise RouteError('Route {route} is not registered.'.format(
                route=repr(route))
            )

        return self._routes[route]

    @staticmethod
    def merged(x, y):
        """ Merges two instances of `Router` into a single one. """

        message = 'Expected {obj} to be of type {router}, but got {type}.'

        if x is not None and not isinstance(x, Router):
            raise TypeError(message.format(
                obj=repr('x'),
                router=Router,
                type=type(x))
            )
        if y is not None and not isinstance(y, Router):
            raise TypeError(message.format(
                obj=repr('y'),
                router=Router,
                type=type(y))
            )

        x = x._routes if x is not None else {}
        y = y._routes if y is not None else {}

        result = {}

        for key, value in chain(x.items(), y.items()):
            if key not in result:
                result[key] = value

        return Router(result)

    @staticmethod
    def new():
        """ Returns a tuple of `(router, router.register)`.

        Usage example:
            class SomeAPI(API):
                __router, route = Router.new()

                @route('a.b.c')
                def _perform(self, ...):
                    ...
        """

        router = Router()

        return router, router.register



from six.moves.urllib.parse import urljoin

from .base import Request


class RelativeRequest(Request):
    """
    Performs requests relative to provided `prefix`.
    """

    def __init__(self, request, prefix):
        self._request = request
        self._prefix = prefix

    def perform(self, method, url, **kwargs):
        url = urljoin(self._prefix, url)

        return self._request.perform(method, url, **kwargs)



from .api.enrich import EnrichAPI
from .api.int import IntAPI
from .api.inspect import InspectAPI
from .api.intel import PrivateIntel, GlobalIntel
from .api.profile import ProfileAPI
from .api.response import ResponseAPI
from .api.commands import CommandsAPI
from .api.sse import SSEDeviceAPI, SSETenantAPI
from .api.user_mgmt import UserMgmtAPI
from .exceptions import CredentialsError
from .request.authorized import ClientAuthorizedRequest, TokenAuthorizedRequest
from .request.logged import LoggedRequest
from .request.proxied import ProxiedRequest
from .request.relative import RelativeRequest
from .request.standard import StandardRequest
from .request.timed import TimedRequest
from .urls import url_for


class ThreatResponse(object):

    def __init__(self, client_id=None, client_password=None,
                 token=None, **options):

        proxy = options.get('proxy')
        timeout = options.get('timeout')
        logger = options.get('logger')
        region = options.get('region')
        environment = options.get('environment')

        request = ProxiedRequest(proxy) if proxy else StandardRequest()
        request = TimedRequest(request, timeout) if timeout else request
        request = LoggedRequest(request, logger) if logger else request
        if token:
            request = TokenAuthorizedRequest(request,
                                             token,
                                             region=region,
                                             environment=environment)
        elif client_id and client_password:
            request = ClientAuthorizedRequest(request,
                                              client_id,
                                              client_password,
                                              region=region,
                                              environment=environment)
        else:
            raise CredentialsError(
                'Credentials must be supplied either '
                'as a pair of client_id and client_password or '
                'as a single token.'
            )

        def request_for(family):
            return RelativeRequest(
                request,
                url_for(region, family, environment)
            )

        self._inspect = InspectAPI(request_for('visibility'))
        self._enrich = EnrichAPI(request_for('visibility'))
        self._int = IntAPI(request_for('visibility'))
        self._response = ResponseAPI(request_for('visibility'))
        self._private_intel = PrivateIntel(request_for('private_intel'))
        self._profile = ProfileAPI(request_for('visibility'))
        self._global_intel = GlobalIntel(request_for('global_intel'))
        self._commands = CommandsAPI(request_for('visibility'))
        self._user_mgmt = UserMgmtAPI(request_for('visibility'))
        self._sse_device = SSEDeviceAPI(request_for('visibility'))
        self._sse_tenant = SSETenantAPI(request_for('visibility'))

    @property
    def inspect(self):
        return self._inspect

    @property
    def enrich(self):
        return self._enrich

    @property
    def int(self):
        return self._int

    @property
    def response(self):
        return self._response

    @property
    def private_intel(self):
        return self._private_intel

    @property
    def profile(self):
        return self._profile

    @property
    def global_intel(self):
        return self._global_intel

    @property
    def commands(self):
        return self._commands

    @property
    def user_mgmt(self):
        return self._user_mgmt

    @property
    def sse_device(self):
        return self._sse_device

    @property
    def sse_tenant(self):
        return self._sse_tenant



from .standard import StandardRequest


class ProxiedRequest(StandardRequest):
    """
    Supports HTTP request proxying via a specified proxy server.
    """

    def __init__(self, proxy):
        super(ProxiedRequest, self).__init__()

        self._proxy = proxy

        self._configure_session_proxies()

    def _configure_session_proxies(self):
        self._session.proxies = {
            'http': self._proxy,
            'https': self._proxy,
        }



from .entity import IntelEntityAPI
from .routing import Router
from .. import urls


class AssetAPI(IntelEntityAPI):

    __router, route = Router.new()

    @route('expire')
    def _perform(self, id_, payload, **kwargs):
        return self._post(
            urls.join(self._url, id_, 'expire'),
            json=payload,
            **kwargs
        )



from .base import API
from .routing import Router
from .. import urls


class UserMgmtAPI(API):
    __router, route = Router.new()

    @route('users.get')
    def _perform(self, user_id, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/user-mgmt/index.html#/User/get_iroh_user_mgmt_users__user_id_
        """

        return self._get(
            urls.join('/iroh/user-mgmt/users', user_id),
            **kwargs
        )

    @route('users.post')
    def _perform(self,
                 user_id,
                 payload,
                 **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/user-mgmt/index.html#/User/post_iroh_user_mgmt_users__user_id_
        """

        return self._post(
            urls.join('/iroh/user-mgmt/users', user_id),
            json=payload,
            **kwargs
        )

    @route('batch.users')
    def _perform(self, user_ids, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/user-mgmt/index.html#/User/get_iroh_user_mgmt_batch_users
        """

        return self._get(
            '/iroh/user-mgmt/batch/users',
            params={'id': user_ids},
            **kwargs
        )

    @route('search.users')
    def _perform(self,
                 payload,
                 sort_by=None,
                 sort_order=None,
                 offset=None,
                 search_after=None,
                 limit=None,
                 **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/user-mgmt/index.html#/User/post_iroh_user_mgmt_search_users
        """

        query = {
            'sort_by': sort_by,
            'sort_order': sort_order,
            'offset': offset,
            'search_after': search_after,
            'limit': limit
        }

        return self._post(
            '/iroh/user-mgmt/search/users',
            json=payload,
            params=query,
            **kwargs
        )



import six

from .base import Request


class LoggedRequest(Request):
    """
    Logs every response of inner request.
    """

    MESSAGE_FORMAT = '{method} {url} {status_code} {reason_phrase}'

    def __init__(self, request, logger):
        self._request = request
        self._logger = logger

    def perform(self, method, url, **kwargs):
        try:
            response = self._request.perform(method, url, **kwargs)
        except Exception:
            self._log_error(method, url)
            raise

        if response.ok:  # 100 <= code < 400.
            self._log_success(method, url, response)
        else:  # 400 <= code < 600.
            self._log_error(method, url, response)

        return response

    @classmethod
    def _format(cls, method, url, response=None):
        return cls.MESSAGE_FORMAT.format(
            method=method.upper(),
            url=url,
            status_code=(
                '' if response is None else
                str(response.status_code)
            ),
            reason_phrase=(
                '' if response is None else
                six.moves.http_client.responses[response.status_code]
            ),
        ).rstrip()

    def _log_success(self, method, url, response):
        message = self._format(method, url, response)

        self._logger.info(message)

    def _log_error(self, method, url, response=None):
        message = self._format(method, url, response)

        if response is None:
            # The same as .error(), but also includes the current traceback.
            self._logger.exception(message)
        else:
            self._logger.error(message)



from .base import API
from .routing import Router
from .. import urls


class EnrichAPI(API):
    __router, route = Router.new()

    @route('health')
    def _perform(self, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-enrich/index.html#/Health/post_iroh_iroh_enrich_health
        """

        return self._post(
            '/iroh/iroh-enrich/health',
            **kwargs
        )

    @route('health.module_instance_id')
    def _perform(self, module_instance_id, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-enrich/index.html#/Health/post_iroh_iroh_enrich_health__module_instance_id_
        """

        return self._post(
            urls.join('/iroh/iroh-enrich/health', module_instance_id),
            **kwargs
        )

    @route('deliberate.observables')
    def _perform(self, payload, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-enrich/index.html#/Deliberate/post_iroh_iroh_enrich_deliberate_observables
        """

        return self._post(
            '/iroh/iroh-enrich/deliberate/observables',
            json=payload,
            **kwargs
        )

    @route('deliberate.sighting')
    def _perform(self, payload, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-enrich/index.html#/Deliberate/post_iroh_iroh_enrich_deliberate_sighting
        """

        return self._post(
            '/iroh/iroh-enrich/deliberate/sighting',
            json=payload,
            **kwargs
        )

    @route('deliberate.sighting_ref')
    def _perform(self, payload, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-enrich/index.html#/Deliberate/post_iroh_iroh_enrich_deliberate_sighting_ref
        """

        return self._post(
            '/iroh/iroh-enrich/deliberate/sighting_ref',
            json=payload,
            **kwargs
        )

    @route('observe.observables')
    def _perform(self, payload, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-enrich/index.html#/Observe/post_iroh_iroh_enrich_observe_observables
        """

        return self._post(
            '/iroh/iroh-enrich/observe/observables',
            json=payload,
            **kwargs
        )

    @route('observe.sighting')
    def _perform(self, payload, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-enrich/index.html#/Observe/post_iroh_iroh_enrich_observe_sighting
        """

        return self._post(
            '/iroh/iroh-enrich/observe/sighting',
            json=payload,
            **kwargs
        )

    @route('observe.sighting_ref')
    def _perform(self, payload, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-enrich/index.html#/Observe/post_iroh_iroh_enrich_observe_sighting_ref
        """

        return self._post(
            '/iroh/iroh-enrich/observe/sighting_ref',
            json=payload,
            **kwargs
        )

    @route('refer.observables')
    def _perform(self, payload, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-enrich/index.html#/Refer/post_iroh_iroh_enrich_refer_observables
        """

        return self._post(
            '/iroh/iroh-enrich/refer/observables',
            json=payload,
            **kwargs
        )

    @route('refer.sighting')
    def _perform(self, payload, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-enrich/index.html#/Refer/post_iroh_iroh_enrich_refer_sighting
        """

        return self._post(
            '/iroh/iroh-enrich/refer/sighting',
            json=payload,
            **kwargs
        )

    @route('refer.sighting_ref')
    def _perform(self, payload, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-enrich/index.html#/Refer/post_iroh_iroh_enrich_refer_sighting_ref
        """

        return self._post(
            '/iroh/iroh-enrich/refer/sighting_ref',
            json=payload,
            **kwargs
        )

    @route('settings.get')
    def _perform(self, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-enrich/index.html#/Settings/get_iroh_iroh_enrich_settings
        """

        return self._get(
            '/iroh/iroh-enrich/settings',
            **kwargs
        )



from .entity import IntelEntityAPI
from .routing import Router
from .. import urls


class FeedAPI(IntelEntityAPI):
    """https://private.intel.amp.cisco.com/index.html#/Feed"""

    __router, route = Router.new()

    def __init__(self, request):
        super(FeedAPI, self).__init__(request, '/ctia/feed')

    @route('view')
    def _perform(self, id_, share_token, **kwargs):
        return self._get(
            urls.join(self._url, id_, 'view'),
            params={'s': share_token},
            **kwargs
        )

    @route('view.txt')
    def _perform(self, id_, share_token, **kwargs):
        return self._get(
            urls.join(self._url, id_, 'view.txt'),
            params={'s': share_token},
            response_type='text',
            **kwargs
        )



import abc

import six


class Request(six.with_metaclass(abc.ABCMeta, object)):
    """
    Interface for performing HTTP requests.
    """

    @abc.abstractmethod
    def perform(self, method, url, **kwargs):
        pass

    def get(self, url, **kwargs):
        return self.perform('GET', url, **kwargs)

    def post(self, url, **kwargs):
        return self.perform('POST', url, **kwargs)

    def put(self, url, **kwargs):
        return self.perform('PUT', url, **kwargs)

    def patch(self, url, **kwargs):
        return self.perform('PATCH', url, **kwargs)

    def delete(self, url, **kwargs):
        return self.perform('DELETE', url, **kwargs)



from .base import API
from .routing import Router


class ProfileAPI(API):
    __router, route = Router.new()

    @route('whoami')
    def _perform(self, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/profile/index.html#/Profile/get_iroh_profile_whoami
        """

        return self._get(
            '/iroh/profile/whoami',
            **kwargs
        )

    @route('org.get')
    def _perform(self, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/profile/index.html#/Profile/get_iroh_profile_org
        """

        return self._get(
            '/iroh/profile/org',
            **kwargs
        )

    @route('org.post')
    def _perform(self, payload, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/profile/index.html#/Profile/post_iroh_profile_org
        """

        return self._post(
            '/iroh/profile/org',
            json=payload,
            **kwargs
        )



import re
import setuptools


def read_version():
    with open('threatresponse/version.py', 'r') as fin:
        return re.search(
            r"^__version__ = '(?P<version>.+)'$",
            fin.read().strip(),
        ).group('version')


def read_readme():
    with open('README.md', 'r') as fin:
        return fin.read().strip()


def read_requirements():
    with open('requirements.txt', 'r') as fin:
        requirements = []
        for line in fin:
            # Discard any comments (i.e. everything after the very first '#').
            line = line.split('#', 1)[0].strip()
            if line:
                requirements.append(line)
        return requirements


NAME = 'threatresponse'

VERSION = read_version()

DESCRIPTION = 'Threat Response API Module'

LONG_DESCRIPTION = read_readme()

LONG_DESCRIPTION_CONTENT_TYPE = 'text/markdown'

URL = 'https://github.com/CiscoSecurity/tr-05-api-module'

AUTHOR = 'Cisco Security'

LICENSE = 'MIT'

PACKAGES = setuptools.find_packages(exclude=['tests', 'tests.*'])

PYTHON_REQUIRES = '>=2.6'

INSTALL_REQUIRES = read_requirements()

KEYWORDS = [
    'cisco', 'security',
    'threat', 'response',
    'api', 'module',
    'python',
]

CLASSIFIERS = [
    'Intended Audience :: Developers',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Topic :: Software Development :: Libraries :: Python Modules',
]


setuptools.setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
    url=URL,
    author=AUTHOR,
    license=LICENSE,
    packages=PACKAGES,
    python_requires=PYTHON_REQUIRES,
    install_requires=INSTALL_REQUIRES,
    keywords=KEYWORDS,
    classifiers=CLASSIFIERS,
)



from .entity import IntelEntityAPI
from .routing import Router
from .. import urls


class SightingAPI(IntelEntityAPI):
    """https://private.intel.amp.cisco.com/index.html#/Sighting"""

    __router, route = Router.new()

    def __init__(self, request):
        super(SightingAPI, self).__init__(request, '/ctia/sighting')

    @route('sightings')
    def _perform(self, observable_type, observable_value, **kwargs):
        return self._get(
            urls.join('/ctia', observable_type, observable_value, 'sightings'),
            **kwargs
        )



from .entity import IntelEntityAPI
from .routing import Router
from .. import urls


class JudgementAPI(IntelEntityAPI):
    """https://private.intel.amp.cisco.com/index.html#/Judgement"""

    __router, route = Router.new()

    def __init__(self, request):
        super(JudgementAPI, self).__init__(request, '/ctia/judgement')

    @route('judgements')
    def _perform(self, observable_type, observable_value, **kwargs):
        return self._get(
            urls.join(
                '/ctia',
                observable_type,
                observable_value,
                'judgements'
            ),
            **kwargs
        )

    @route('expire')
    def _perform(self, id_, payload, **kwargs):
        return self._post(
            urls.join(self._url, id_, 'expire'),
            json=payload,
            **kwargs
        )



from .base import API
from .routing import Router


class InspectAPI(API):
    __router, route = Router.new()

    @route('inspect')
    def _perform(self, payload, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-inspect/index.html#/INSPECT/post_iroh_iroh_inspect_inspect
        """

        return self._post(
            '/iroh/iroh-inspect/inspect',
            json=payload,
            **kwargs
        )



from .base import API
from .routing import Router
from .. import urls


class ResponseAPI(API):
    __router, route = Router.new()

    @route('respond.observables')
    def _perform(self, payload, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-response/index.html#/Response/post_iroh_iroh_response_respond_observables
        """

        return self._post(
            '/iroh/iroh-response/respond/observables',
            json=payload,
            **kwargs
        )

    @route('respond.sighting')
    def _perform(self, payload, **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-response/index.html#/Response/post_iroh_iroh_response_respond_sighting
        """

        return self._post(
            '/iroh/iroh-response/respond/sighting',
            json=payload,
            **kwargs
        )

    @route('respond.trigger')
    def _perform(self,
                 module_name,
                 action_id,
                 observable_type=None,
                 observable_value=None,
                 **kwargs):
        """
        https://visibility.amp.cisco.com/iroh/iroh-response/index.html#/Response/post_iroh_iroh_response_respond_trigger__module_name___action_id_
        """

        url = urls.join(
            '/iroh/iroh-response/respond/trigger',
            module_name,
            action_id
        )

        # Extend optional module-specific query params with the required ones.
        query = kwargs.pop('params', {})
        if observable_type and observable_value:
            query.update({
                'observable_type': observable_type,
                'observable_value': observable_value,
            })

        return self._post(url, params=query, **kwargs)



from copy import deepcopy

from .base import API
from .routing import Router


class CommandsAPI(API):
    __router, route = Router.new()

    @route('verdict')
    def _perform(self, payload, **kwargs):
        """
        Command allow to simple query CTR
        for a verdict for a bunch of observables
        """

        response = self._post(
            '/iroh/iroh-inspect/inspect',
            json={'content': str(payload)},
            **kwargs
        )

        response = self._post(
            '/iroh/iroh-enrich/deliberate/observables',
            json=response,
            **kwargs
        )
        verdicts = build_array_for_verdicts(response)
        return {"response": response, "verdicts": verdicts}

    @route('targets')
    def _perform(self, payload, **kwargs):
        """
        Command allow to simple query CTR for a targets
        for a bunch of observables
        """

        response = self._post(
            '/iroh/iroh-inspect/inspect',
            json={'content': str(payload)},
            **kwargs
        )

        response = self._post(
            '/iroh/iroh-enrich/observe/observables',
            json=response,
            **kwargs
        )

        result = build_array_for_targets(response)

        return {"response": response, "targets": result}


def build_array_for_verdicts(verdict_dict):
    verdicts = []

    # According to the official documentation, `disposition_name` is
    # optional, so simply infer it from required `disposition`.
    disposition_map = {
        1: 'Clean',
        2: 'Malicious',
        3: 'Suspicious',
        4: 'Common',
        5: 'Unknown',
    }

    for module in verdict_dict.get('data', []):
        module_name = module['module']
        module_type_id = module['module_type_id']
        module_instance_id = module['module_instance_id']

        for doc in module.get('data', {}) \
                .get('verdicts', {}) \
                .get('docs', []):
            verdicts.append({
                'observable_value': doc['observable']['value'],
                'observable_type': doc['observable']['type'],
                'expiration': doc['valid_time'].get('end_time', ''),
                'module': module_name,
                'module_type_id': module_type_id,
                'module_instance_id': module_instance_id,
                'disposition_name': disposition_map[doc['disposition']],
            })

    return verdicts


def build_array_for_targets(targets_dict):
    result = []

    for module in targets_dict.get('data', []):
        module_name = module['module']
        module_type_id = module['module_type_id']
        module_instance_id = module['module_instance_id']
        targets = []

        for doc in module.get('data', {}) \
                .get('sightings', {}) \
                .get('docs', []):

            for target in doc.get('targets', []):
                element = deepcopy(target)
                element.pop('observed_time', None)
                if element not in targets:
                    targets.append(element)

        result.append({
            'module': module_name,
            'module_type_id': module_type_id,
            'module_instance_id': module_instance_id,
            'targets': targets
        })
    return result



class Resolution(object):
    """ Represents a resolution of attribute chains.
    `resolution.x.y.z` would contain `resolution._route = ['x', 'y', 'z']`.
    """

    def __init__(self, owner, router, route=None):
        self._owner = owner
        self._router = router
        self._route = route or []

    def __call__(self, *args, **kwargs):
        """ Invokes a method by the built route. """

        route = '.'.join(self._route)
        method = self._router.resolve(route)

        return method(self._owner, *args, **kwargs)

    def __getattr__(self, item):
        return Resolution(self._owner, self._router, self._route + [item])



from .base import Request


class TimedRequest(Request):
    """
    Sets the default request timeout (unless explicitly specified).
    """

    def __init__(self, request, timeout):
        self._request = request
        self._timeout = timeout

    def perform(self, method, url, **kwargs):
        kwargs.setdefault('timeout', self._timeout)

        return self._request.perform(method, url, **kwargs)
